cdef class _BanditWrapper:
    cdef _Bandit *_handle

    def __cinit__(self):
        self._handle = NULL

    def __dealloc__(self):
        if self._handle != NULL:
            del self._handle

    def _init(self, config):
        self._handle = new _Bandit(config)
        typ, ver = b'bandit', 1
        return (
            lambda: self._handle.get_config().decode('utf8'),
            lambda: self._handle.dump(typ, ver),
            lambda x: self._handle.load(x, typ, ver),
            lambda: self._handle.clear(),
            typ,
        )

    def register_arm(self, arm_id):
        return self._handle.register_arm(arm_id.encode('utf8'))

    def delete_arm(self, arm_id):
        return self._handle.delete_arm(arm_id.encode('utf8'))

    def select_arm(self, player_id):
        return self._handle.select_arm(player_id.encode('utf8')).decode('utf8')

    def register_reward(self, player_id, arm_id, reward):
        return self._handle.register_reward(
            player_id.encode('utf8'), arm_id.encode('utf8'), reward)

    def get_arm_info(self, player_id):
        cdef map[string, arm_info] r
        r = self._handle.get_arm_info(player_id.encode('utf8'))
        return {
            it.first.decode('utf8'): ArmInfo(it.second.trial_count, it.second.weight)
            for it in r
        }

    def reset(self, player_id):
        return self._handle.reset(player_id.encode('utf8'))



import json
import math
import random
import unittest

from embedded_jubatus import Stat


class TestStat(unittest.TestCase):
    def test_invalid_configs(self):
        self.assertRaises(TypeError, Stat)
        self.assertRaises(RuntimeError, Stat, {})
        self.assertRaises(RuntimeError, Stat, {'window_size': 'hoge'})

    def test(self):
        window_size = 2000
        config = {
            'window_size': window_size
        }
        s = Stat(config)

        x_avg, x_stddev = 0.0, 1.0
        y_avg, y_stddev = 10.0, 10.0
        x_list, y_list = [], []

        for _ in range(window_size // 2):
            xv = random.gauss(x_avg, x_stddev)
            yv = random.gauss(y_avg, y_stddev)
            self.assertTrue(s.push('x', xv))
            self.assertTrue(s.push('y', yv))
            x_list.append(xv)
            y_list.append(yv)

        x_sum, y_sum = sum(x_list), sum(y_list)
        x_avg, y_avg = x_sum / len(x_list), y_sum / len(y_list)
        self.assertAlmostEqual(x_sum, s.sum('x'))
        self.assertAlmostEqual(y_sum, s.sum('y'))
        self.assertAlmostEqual(
            math.sqrt(sum([(x - x_avg)**2 for x in x_list]) / len(x_list)),
            s.stddev('x'))
        self.assertAlmostEqual(
            math.sqrt(sum([(y - y_avg)**2 for y in y_list]) / len(y_list)),
            s.stddev('y'))
        self.assertAlmostEqual(max(x_list), s.max('x'))
        self.assertAlmostEqual(max(y_list), s.max('y'))
        self.assertAlmostEqual(min(x_list), s.min('x'))
        self.assertAlmostEqual(min(y_list), s.min('y'))
        self.assertTrue(isinstance(s.entropy(''), float))
        self.assertTrue(isinstance(s.moment('x', 1, 1.0), float))

        model = s.save_bytes()
        s = Stat(config)
        s.load_bytes(model)
        self.assertEqual(config, json.loads(s.get_config()))
        self.assertAlmostEqual(x_sum, s.sum('x'))



cdef datum_py2native(pd, datum& d):
    d.string_values_.clear()
    for k, v in pd.string_values:
        k = k.encode('utf8')
        v = v.encode('utf8')
        d.string_values_.push_back(pair[string, string](k, v))
    d.num_values_.clear()
    for k, v in pd.num_values:
        k = k.encode('utf8')
        d.num_values_.push_back(pair[string, double](k, v))
    d.binary_values_.clear()
    for k, v in pd.binary_values:
        k = k.encode('utf8')
        d.binary_values_.push_back(pair[string, string](k, v))

cdef datum_native2py(datum& d):
    ret = Datum()
    for i in range(d.string_values_.size()):
        k = d.string_values_[i].first.decode('utf8')
        v = d.string_values_[i].second.decode('utf8')
        ret.add_string(k, v)
    for i in range(d.num_values_.size()):
        k = d.num_values_[i].first.decode('utf8')
        v = d.num_values_[i].second
        ret.add_number(k, v)
    for i in range(d.binary_values_.size()):
        k = d.binary_values_[i].first.decode('utf8')
        v = d.binary_values_[i].second
        ret.add_binary(k, v)
    return ret

IF NUMPY:

    cdef ndarray_to_datum(c_np.ndarray[c_np.float64_t, ndim=2] X, int i, datum& d, vector[string]& cache):
        d.string_values_.clear()
        d.num_values_.clear()
        d.binary_values_.clear()

        cdef int j
        for j in range(cache.size(), X.shape[1]):
            cache.push_back(lexical_cast[string, int](j))
        for j in range(X.shape[1]):
            if X[i, j] != 0.0:
                d.num_values_.push_back(pair[string, double](cache[j], X[i, j]))

    cdef csr_to_datum(c_np.ndarray[c_np.float64_t, ndim=1] data,
                      c_np.ndarray[c_np.int32_t, ndim=1] indices,
                      c_np.ndarray[c_np.int32_t, ndim=1] indptr,
                      int i, datum& d, vector[string]& cache):
        d.string_values_.clear()
        d.num_values_.clear()
        d.binary_values_.clear()

        cdef int j = indptr[i]
        cdef int k = indptr[i + 1]
        cdef int l, m
        for l in range(j, k):
            for m in range(cache.size(), indices[l] + 1):
                cache.push_back(lexical_cast[string, int](m))
            d.num_values_.push_back(pair[string, double](cache[indices[l]], data[l]))

ELSE:

    cdef ndarray_to_datum(X, int i, datum& d, vector[string]& cache):
        raise RuntimeError

    cdef csr_to_datum(data, indices, indptr, int i, datum& d, vector[string]& cache):
        raise RuntimeError

cdef props_py2native(p, prop_t& out):
    for k, v in p.items():
        out.insert(pair[string, string](k.encode('utf8'), v.encode('utf8')))

cdef props_native2py(prop_t& p):
    r = {}
    for it in p:
        r[it.first.decode('utf8')] = it.second.decode('utf8')
    return r

cdef edges_native2py(const vector[edge_id_t]& edges):
    ret = []
    for i in range(edges.size()):
        ret.append(edges[i])
    return ret

cdef preset_query_py2native(query, preset_query& q):
    for x in query.edge_query:
        q.edge_query.push_back(pair[string, string](
            x.from_id.encode('ascii'), x.to_id.encode('ascii')))
    for x in query.node_query:
        q.node_query.push_back(pair[string, string](
            x.from_id.encode('ascii'), x.to_id.encode('ascii')))



from libc.stdint cimport uint64_t
from libcpp cimport bool
from libcpp.string cimport string
from libcpp.vector cimport vector
from libcpp.pair cimport pair
from libcpp.map cimport map

ctypedef vector[pair[string, float]] sfv_t
ctypedef map[string, string] prop_t
ctypedef uint64_t edge_id_t
ctypedef uint64_t node_id_t

cdef extern from '_wrapper.h' nogil:
    cdef cppclass _Classifier:
        _Classifier(const string& config) except +
        void train(const string& config, const datum& d)
        vector[classify_result_elem] classify(const datum& d)
        vector[pair[string, uint64_t]] get_labels()
        bool set_label(const string& new_label)
        bool delete_label(const string& target_label)
        string dump(const string& type, uint64_t ver)
        void load(const string& data, const string& type, uint64_t ver)
        string get_config()
        void clear()

    cdef cppclass _Regression:
        _Regression(const string& config) except +
        void train(float score, const datum& d)
        float estimate(const datum& d)
        string dump(const string& type, uint64_t ver)
        void load(const string& data, const string& type, uint64_t ver)
        string get_config()
        void clear()

    cdef cppclass _Recommender:
        _Recommender(const string& config) except +
        void clear_row(const string& id)
        void update_row(const string& id, const datum& d)
        datum complete_row_from_id(const string& id)
        datum complete_row_from_datum(const datum& d)
        vector[pair[string, float]] similar_row_from_id(const string& id, size_t ret_num)
        vector[pair[string, float]] similar_row_from_datum(const datum& d, size_t ret_num)
        datum decode_row(const string& id)
        vector[string] get_all_rows()
        float calc_similarity(const datum& l, const datum& r)
        float calc_l2norm(const datum& d)
        string dump(const string& type, uint64_t ver)
        void load(const string& data, const string& type, uint64_t ver)
        string get_config()
        void clear()

    cdef cppclass _NearestNeighbor:
        _NearestNeighbor(const string& config) except +
        void set_row(const string& id, const datum& d)
        vector[pair[string, float]] neighbor_row_from_id(const string& id, size_t size)
        vector[pair[string, float]] neighbor_row_from_datum(const datum& d, size_t size)
        vector[pair[string, float]] similar_row_from_id(const string& id, size_t size)
        vector[pair[string, float]] similar_row_from_datum(const datum& d, size_t size)
        vector[string] get_all_rows()
        string dump(const string& type, uint64_t ver)
        void load(const string& data, const string& type, uint64_t ver)
        string get_config()
        void clear()

    cdef cppclass _Anomaly:
        _Anomaly(const string& config) except +
        void clear_row(const string& id)
        pair[string, float] add(const datum& d)
        float update(const string& id, const datum& d)
        float overwrite(const string& id, const datum& d)
        float calc_score(const datum& d)
        vector[string] get_all_rows()
        string dump(const string& type, uint64_t ver)
        void load(const string& data, const string& type, uint64_t ver)
        string get_config()
        void clear()

    cdef cppclass _Clustering:
        _Clustering(const string& config) except +
        void push(const vector[indexed_point]& points)
        size_t get_revision()
        vector[vector[pair[double, datum]]] get_core_members()
        vector[datum] get_k_center()
        datum get_nearest_center(const datum& d)
        vector[pair[double, datum]] get_nearest_members(const datum& d)
        vector[vector[pair[double, string]]] get_core_members_light()
        vector[pair[double, string]] get_nearest_members_light(const datum& d)
        string dump(const string& type, uint64_t ver)
        void load(const string& data, const string& type, uint64_t ver)
        string get_config()
        void clear()

    cdef cppclass _Burst:
        cppclass Batch:
            int all_data_count
            int relevant_data_count
            double burst_weight
        _Burst(const string& config) except +
        bool add_document(const string& str, double pos)
        pair[double,vector[Batch]] get_result(const string& keyword)
        pair[double,vector[Batch]] get_result_at(const string& keyword, double pos)
        map[string, pair[double,vector[Batch]]] get_all_bursted_results()
        map[string, pair[double,vector[Batch]]] get_all_bursted_results_at(double pos)
        vector[keyword_with_params] get_all_keywords()
        bool add_keyword(const string& keyword, const keyword_params& params)
        bool remove_keyword(const string& keyword)
        bool remove_all_keywords()
        void calculate_results()
        string dump(const string& type, uint64_t ver)
        void load(const string& data, const string& type, uint64_t ver)
        string get_config()
        void clear()

    cdef cppclass _Bandit:
        _Bandit(const string& config) except +
        bool register_arm(const string& arm_id)
        bool delete_arm(const string& arm_id)
        string select_arm(const string& player_id)
        bool register_reward(const string& player_id, const string& arm_id, double reward)
        map[string, arm_info] get_arm_info(const string& player_id)
        bool reset(const string& player_id)
        string dump(const string& type, uint64_t ver)
        void load(const string& data, const string& type, uint64_t ver)
        string get_config()
        void clear()

    cdef cppclass _Stat:
        _Stat(const string& config) except +
        void push(const string& key, double value)
        double sum(const string& key)
        double stddev(const string& key)
        double max(const string& key)
        double min(const string& key)
        double entropy()
        double moment(const string& key, int degree, double center)
        string dump(const string& type, uint64_t ver)
        void load(const string& data, const string& type, uint64_t ver)
        string get_config()
        void clear()

    cdef cppclass _Weight:
        _Weight(const string& config) except +
        sfv_t update(const datum& d)
        sfv_t calc_weight(const datum& d)
        string dump(const string& type, uint64_t ver)
        void load(const string& data, const string& type, uint64_t ver)
        string get_config()
        void clear()

    cdef cppclass _Graph:
        _Graph(const string& config) except +
        string create_node()
        bool remove_node(const string& node_id)
        bool update_node(const string& node_id,
                         const prop_t& properties)
        edge_id_t create_edge(const string& src,
                              const string& target,
                              const prop_t& properties)
        bool update_edge(edge_id_t edge_id,
                         const prop_t& properties)
        void remove_edge(edge_id_t edge_id)

        double get_centrality(const string& node_id,
                              int centrality_type,
                              const preset_query& q)
        void add_centrality_query(const preset_query& q)
        void add_shortest_path_query(const preset_query& q)
        void remove_centrality_query(const preset_query& q)
        void remove_shortest_path_query(const preset_query& q)
        vector[node_id_t] get_shortest_path(const string& src,
                                            const string& target,
                                            uint64_t max_hop,
                                            const preset_query& q)

        void update_index()
        node_info get_node(const string& node_id)
        edge_info get_edge(edge_id_t eid)

        string dump(const string& type, uint64_t ver)
        void load(const string& data, const string& type, uint64_t ver)
        string get_config()
        void clear()

cdef extern from 'jubatus/core/fv_converter/datum.hpp' namespace 'jubatus::core::fv_converter' nogil:
    cdef cppclass datum:
        vector[pair[string, string]] string_values_
        vector[pair[string, double]] num_values_
        vector[pair[string, string]] binary_values_

cdef extern from 'jubatus/core/classifier/classifier_type.hpp' namespace 'jubatus::core::classifier' nogil:
    cdef cppclass classify_result_elem:
        string label
        float score

cdef extern from 'jubatus/core/burst/burst.hpp' namespace 'jubatus::core::burst' nogil:
    cdef cppclass keyword_params:
        double scaling_param
        double gamma

    cdef cppclass keyword_with_params:
        string keyword
        double scaling_param
        double gamma

cdef extern from 'jubatus/core/bandit/arm_info.hpp' namespace 'jubatus::core::bandit' nogil:
    cdef cppclass arm_info:
        int trial_count
        double weight

cdef extern from 'jubatus/util/lang/cast.h' namespace 'jubatus::util::lang' nogil:
    cdef T lexical_cast[T, S](const S& arg)

cdef extern from 'jubatus/core/graph/graph_type.hpp' namespace 'jubatus::core::graph' nogil:
    cdef cppclass node_info:
        prop_t property
        vector[edge_id_t] in_edges
        vector[edge_id_t] out_edges

    cdef cppclass edge_info:
        prop_t p
        node_id_t src
        node_id_t tgt

    cdef cppclass preset_query:
        vector[pair[string, string]] edge_query
        vector[pair[string, string]] node_query

cdef extern from 'jubatus/core/clustering/types.hpp' namespace 'jubatus::core::clustering' nogil:
    cdef cppclass indexed_point:
        string id
        datum point



cdef class _ClassifierWrapper:
    cdef _Classifier *_handle

    def __cinit__(self):
        self._handle = NULL

    def __dealloc__(self):
        if self._handle != NULL:
            del self._handle

    def _init(self, config):
        self._handle = new _Classifier(config)
        self.classes_ = None
        typ, ver = b'classifier', 1
        return (
            lambda: self._handle.get_config().decode('utf8'),
            lambda: self._handle.dump(typ, ver),
            lambda x: self._handle.load(x, typ, ver),
            lambda: self._handle.clear(),
            typ,
        )

    def train(self, data):
        cdef string l
        cdef datum d
        if isinstance(data, list):
            if len(data) == 0:
                return 0
            if isinstance(data[0], LabeledDatum):
                for r in data:
                    label, datum = r.label, r.data
                    l = label.encode('utf8')
                    datum_py2native(datum, d)
                    self._handle.train(l, d)
            elif isinstance(data[0], (tuple, list)) and len(data[0]) == 2:
                for label, datum in data:
                    l = label.encode('utf8')
                    datum_py2native(datum, d)
                    self._handle.train(l, d)
            else:
                raise ValueError
            return len(data)
        else:
            raise ValueError

    def classify(self, data):
        cdef datum d
        cdef vector[classify_result_elem] r
        if isinstance(data, list):
            ret = []
            for x in data:
                datum_py2native(x, d)
                r = self._handle.classify(d)
                ret.append([
                    EstimateResult(r[i].label.decode('utf8'), r[i].score)
                    for i in range(r.size())
                ])
            return ret
        else:
            raise ValueError

    def get_labels(self):
        cdef vector[pair[string, uint64_t]] ret = self._handle.get_labels()
        return {ret[i].first.decode('utf8'): ret[i].second for i in range(ret.size())}

    def set_label(self, new_label):
        return self._handle.set_label(new_label.encode('utf8'))

    def delete_label(self, target_label):
        return self._handle.delete_label(target_label.encode('utf8'))

    def fit(self, X, y):
        self._handle.clear()
        self.classes_ = None
        return self.partial_fit(X, y)

    def partial_fit(self, X, y):
        import numpy as np
        if len(X.shape) != 2 or len(y.shape) != 1 or X.shape[0] != y.shape[0]:
            raise ValueError('invalid shape')
        cdef int i, j, max_label
        cdef rows = X.shape[0]
        cdef datum d
        cdef vector[string] cache
        cdef int is_ndarray = isinstance(X, np.ndarray)
        cdef int is_csr = (type(X).__name__ == 'csr_matrix')
        if not (is_ndarray or is_csr):
            raise ValueError
        if self.classes_ is None:
            self.classes_ = []
        max_label = len(self.classes_) - 1
        for i in range(rows):
            if is_ndarray:
                ndarray_to_datum(X, i, d, cache)
            else:
                csr_to_datum(X.data, X.indices, X.indptr, i, d, cache)
            for j in range(cache.size(), y[i] + 1):
                cache.push_back(lexical_cast[string, int](j))
            if max_label < y[i]:
                max_label = y[i]
            self._handle.train(cache[y[i]], d)
        for j in range(len(self.classes_), max_label + 1):
            self.classes_.append(j)
        return self

    def decision_function(self, X):
        import numpy as np
        if len(X.shape) != 2:
            raise ValueError('invalid X.shape')
        cdef int is_ndarray = isinstance(X, np.ndarray)
        cdef int is_csr = (type(X).__name__ == 'csr_matrix')
        if not (is_ndarray or is_csr):
            raise ValueError
        cdef int i, j, k
        cdef double score
        cdef datum d
        cdef vector[string] cache
        cdef vector[classify_result_elem] r
        cdef int rows = X.shape[0]
        ret = None
        for i in range(rows):
            if is_ndarray:
                ndarray_to_datum(X, i, d, cache)
            else:
                csr_to_datum(X.data, X.indices, X.indptr, i, d, cache)
            r = self._handle.classify(d)
            if r.size() == 0:
                return np.zeros(rows)
            if r.size() == 2:
                if ret is None:
                    ret = np.zeros(rows)
                if r[0].score < r[1].score:
                    j = lexical_cast[int, string](r[1].label)
                    score = r[1].score
                else:
                    j = lexical_cast[int, string](r[0].label)
                    score = r[0].score
                if j == 0:
                    ret[i] = -score
                else:
                    ret[i] = score
            else:
                if ret is None:
                    ret = np.zeros((rows, r.size()))
                for j in range(r.size()):
                    k = lexical_cast[int, string](r[j].label)
                    ret[i,k] = r[j].score
        return ret

    def predict(self, X):
        import numpy as np
        if len(X.shape) != 2:
            raise ValueError('invalid X.shape')
        cdef int is_ndarray = isinstance(X, np.ndarray)
        cdef int is_csr = (type(X).__name__ == 'csr_matrix')
        if not (is_ndarray or is_csr):
            raise ValueError
        cdef int i, j, max_j
        cdef double max_score
        cdef datum d
        cdef vector[string] cache
        cdef vector[classify_result_elem] r
        cdef int rows = X.shape[0]
        IF NUMPY:
            cdef c_np.ndarray[c_np.int32_t, ndim=1] ret = np.zeros(rows, dtype=np.int32)
        ELSE:
            ret = np.zeros(rows, dtype=np.int32)
        for i in range(rows):
            if is_ndarray:
                ndarray_to_datum(X, i, d, cache)
            else:
                csr_to_datum(X.data, X.indices, X.indptr, i, d, cache)
            r = self._handle.classify(d)
            if r.size() == 0:
                break
            max_j, max_score = 0, r[0].score
            for j in range(1, r.size()):
                if r[j].score > max_score:
                    max_j, max_score = j, r[j].score
            ret[i] = lexical_cast[int, string](r[max_j].label)
        return ret



import json
import unittest

from embedded_jubatus import Recommender
from jubatus.common import Datum
from jubatus.recommender.types import IdWithScore


CONFIG = {
    'method': 'lsh',
    'parameter': {
        'hash_num': 512,
    },
    'converter': {
        'num_filter_types': {},
        'num_filter_rules': [],
        'string_filter_types': {},
        'string_filter_rules': [],
        'num_types': {},
        'num_rules': [
            {'key': '*', 'type': 'num'}
        ],
        'string_types': {},
        'string_rules': [
            {'key': '*', 'type': 'space',
             'sample_weight': 'bin', 'global_weight': 'bin'}
        ]
    },
}


class TestRecommender(unittest.TestCase):
    def test_invalid_config(self):
        self.assertRaises(TypeError, Recommender)
        self.assertRaises(ValueError, Recommender, {})
        self.assertRaises(TypeError, Recommender, {'method': 'hoge'})
        self.assertRaises(RuntimeError, Recommender,
                          {'method': 'hoge', 'converter': {}})
        invalid_config = dict(CONFIG)
        invalid_config['method'] = 'hoge'
        self.assertRaises(RuntimeError, Recommender, invalid_config)

    def test(self):
        def _valid_result(r):
            self.assertTrue(isinstance(r, Datum))
            d = dict(r.num_values)
            self.assertTrue(
                d.get('x', None) is not None and d.get('y', None) is not None)

        x = Recommender(CONFIG)
        x.update_row('0', Datum({'x': 0.9, 'y': 4.9}))
        x.update_row('1', Datum({'x': 1, 'y': 5}))
        x.update_row('2', Datum({'x': 1.15, 'y': 5.15}))
        x.update_row('3', Datum({'x': 1.2, 'y': 5.1}))
        x.update_row('4', Datum({'x': 1.05}))
        _valid_result(x.complete_row_from_datum(Datum({'x': 1.1})))
        _valid_result(x.complete_row_from_id('4'))
        r = x.similar_row_from_id('2', 3)
        self.assertTrue(isinstance(r, list))
        self.assertTrue(isinstance(r[0], IdWithScore))
        r = x.similar_row_from_datum(Datum({'y': 5.05}), 3)
        self.assertTrue(isinstance(r, list))
        self.assertTrue(isinstance(r[0], IdWithScore))
        _valid_result(x.decode_row('0'))
        assert set(x.get_all_rows()) == set([str(i) for i in range(5)])
        r = x.calc_similarity(Datum({'x': 1}), Datum({'y': 5}))
        self.assertTrue(isinstance(r, float))
        r = x.calc_l2norm(Datum({'x': 1, 'y': 5}))
        self.assertTrue(isinstance(r, float))

        model = x.save_bytes()
        x.clear()
        self.assertEqual([], x.get_all_rows())
        self.assertEqual(CONFIG, json.loads(x.get_config()))

        x = Recommender(CONFIG)
        x.load_bytes(model)
        self.assertTrue(x.get_all_rows())
        _valid_result(x.complete_row_from_id('4'))



cdef class _RecommenderWrapper:
    cdef _Recommender *_handle

    def __cinit__(self):
        self._handle = NULL

    def __dealloc__(self):
        if self._handle != NULL:
            del self._handle

    def _init(self, config):
        self._handle = new _Recommender(config)
        typ, ver = b'recommender', 1
        return (
            lambda: self._handle.get_config().decode('utf8'),
            lambda: self._handle.dump(typ, ver),
            lambda x: self._handle.load(x, typ, ver),
            lambda: self._handle.clear(),
            typ,
        )

    def clear_row(self, id_):
        self._handle.clear_row(id_.encode('utf8'))
        return True

    def update_row(self, id_, row):
        cdef datum d
        datum_py2native(row, d)
        self._handle.update_row(id_.encode('utf8'), d)
        return True

    def complete_row_from_id(self, id_):
        cdef datum d = self._handle.complete_row_from_id(id_.encode('utf8'))
        return datum_native2py(d)

    def complete_row_from_datum(self, row):
        cdef datum d0
        datum_py2native(row, d0)
        cdef datum d1 = self._handle.complete_row_from_datum(d0)
        return datum_native2py(d1)

    def similar_row_from_id(self, id_, size):
        cdef vector[pair[string, float]] ret
        ret = self._handle.similar_row_from_id(id_.encode('utf8'), size)
        return [
            RecommenderIdWithScore(ret[i].first.decode('utf8'), ret[i].second)
            for i in range(ret.size())
        ]

    def similar_row_from_datum(self, row, size):
        cdef vector[pair[string, float]] ret
        cdef datum d
        datum_py2native(row, d)
        ret = self._handle.similar_row_from_datum(d, size)
        return [
            RecommenderIdWithScore(ret[i].first.decode('utf8'), ret[i].second)
            for i in range(ret.size())
        ]

    def decode_row(self, id_):
        cdef datum d = self._handle.decode_row(id_.encode('utf8'))
        return datum_native2py(d)

    def get_all_rows(self):
        cdef vector[string] ret = self._handle.get_all_rows()
        return [(<string>ret[i]).decode('utf8') for i in range(ret.size())]

    def calc_similarity(self, l, r):
        cdef datum d0
        cdef datum d1
        datum_py2native(l, d0)
        datum_py2native(r, d1)
        return self._handle.calc_similarity(d0, d1)

    def calc_l2norm(self, row):
        cdef datum d
        datum_py2native(row, d)
        return self._handle.calc_l2norm(d)



cdef class _RegressionWrapper:
    cdef _Regression *_handle

    def __cinit__(self):
        self._handle = NULL

    def __dealloc__(self):
        if self._handle != NULL:
            del self._handle

    def _init(self, config):
        self._handle = new _Regression(config)
        typ, ver = b'regression', 1
        return (
            lambda: self._handle.get_config().decode('utf8'),
            lambda: self._handle.dump(typ, ver),
            lambda x: self._handle.load(x, typ, ver),
            lambda: self._handle.clear(),
            typ,
        )

    def train(self, data):
        cdef datum d
        if isinstance(data, list):
            if len(data) == 0:
                return 0
            if isinstance(data[0], ScoredDatum):
                for r in data:
                    score = float(r.score)
                    datum_py2native(r.data, d)
                    self._handle.train(score, d)
            elif isinstance(data[0], (tuple, list)):
                for score, datum in data:
                    datum_py2native(datum, d)
                    self._handle.train(float(score), d)
            else:
                raise ValueError
            return len(data)
        else:
            raise ValueError

    def estimate(self, data):
        cdef datum d
        if isinstance(data, list):
            ret = []
            for datum in data:
                datum_py2native(datum, d)
                ret.append(self._handle.estimate(d))
            return ret
        else:
            raise ValueError



import json
import unittest

from embedded_jubatus import Burst
from jubatus.burst.types import Batch
from jubatus.burst.types import Document
from jubatus.burst.types import KeywordWithParams
from jubatus.burst.types import Window


CONFIG = {
    'parameter': {
        'window_batch_size': 5,
        'batch_interval': 10,
        'max_reuse_batch_num': 5,
        'costcut_threshold': -1,
        'result_window_rotate_size': 5
    },
    'method': 'burst',
}


class TestBurst(unittest.TestCase):
    def test_invalid_configs(self):
        self.assertRaises(TypeError, Burst)
        self.assertRaises(ValueError, Burst, {})
        self.assertRaises(RuntimeError, Burst, {'method': 'hoge'})
        invalid_config = dict(CONFIG)
        invalid_config['parameter'] = {'hoge': 0.1}
        self.assertRaises(RuntimeError, Burst, invalid_config)

    def test(self):
        x = Burst(CONFIG)
        balse_kwp = KeywordWithParams('balse', 1.001, 0.1)
        self.assertTrue(x.add_keyword(balse_kwp))
        k = x.get_all_keywords()
        self.assertEqual(1, len(k))
        self.assertTrue(isinstance(k[0], KeywordWithParams))
        self.assertEqual('balse', k[0].keyword)
        self.assertTrue(x.remove_keyword('balse'))
        self.assertFalse(x.remove_keyword('hoge'))
        self.assertEqual([], x.get_all_keywords())
        x.add_keyword(balse_kwp)
        self.assertTrue(x.remove_all_keywords())
        self.assertEqual([], x.get_all_keywords())
        x.add_keyword(balse_kwp)

        def add(pos, burst_count, nonburst_count):
            pos = float(pos)
            x.add_documents(
                [Document(pos, 'balse')] * burst_count +
                [Document(pos, 'jubatus')] * nonburst_count
            )

        # Time   Burst  Non-Burst
        add(1,       5,      30)
        add(11,     15,      50)
        add(21,    500,      10)
        add(31,   2000,      10)
        add(41,  22222,      40)
        add(51,     10,      10)
        add(61,      5,      25)

        ret = x.get_result('balse')
        self.assertTrue(isinstance(ret, Window))
        self.assertTrue(isinstance(ret.batches[0], Batch))
        ret = x.get_result_at('balse', 41)
        self.assertTrue(isinstance(ret, Window))
        self.assertTrue(isinstance(ret.batches[0], Batch))

        ret = x.get_all_bursted_results_at(41)
        self.assertTrue(ret, dict)
        self.assertEqual(['balse'], list(ret.keys()))
        self.assertTrue(isinstance(ret['balse'], Window))
        self.assertTrue(isinstance(ret['balse'].batches[0], Batch))

        model = x.save_bytes()
        x = Burst(CONFIG)
        x.load_bytes(model)
        self.assertEqual(CONFIG, json.loads(x.get_config()))
        self.assertEqual(1, len(x.get_all_keywords()))
        ret = x.get_result_at('balse', 41)
        self.assertTrue(isinstance(ret, Window))
        self.assertTrue(isinstance(ret.batches[0], Batch))



import os.path
import platform

from Cython.Build import cythonize
from setuptools import Extension
from setuptools import setup

def read(name):
    with open(os.path.join(os.path.dirname(__file__), name)) as f:
        return f.read()

py_defines = []
include_dirs = []
try:
    import numpy
    py_defines.append(('NUMPY', True))
    include_dirs.append(numpy.get_include())
except ImportError:
    py_defines.append(('NUMPY', False))

defs_path = 'src/defs.pyx'
defs = '\n'.join(['DEF ' + n + ' = ' + str(v) for n, v in py_defines])
old_defs = open(defs_path, 'r').read() if os.path.isfile(defs_path) else ''
if defs != old_defs:
    with open('src/defs.pyx', 'w') as f:
        f.write(defs)

if platform.system() == 'Darwin':
    extra_compile_args = ['-std=c++11']
else:
    extra_compile_args = []

setup(
    name='embedded_jubatus',
    version=read('VERSION').rstrip(),
    description='embedded-jubatus-python is a Python bridge to call Jubatus Core library.',
    long_description=read('README.rst'),
    author='PFN & NTT',
    author_email='jubatus-team@googlegroups.com',
    url='http://jubat.us',
    download_url='http://pypi.python.org/pypi/embedded_jubatus',
    license='LGPLv2',
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],
    ext_modules=cythonize([
        Extension(
            'embedded_jubatus',
            [
                'src/embedded_jubatus.pyx',
                'src/_wrapper.cpp',
                'src/_model.cpp'
            ],
            include_dirs=include_dirs,
            libraries=['jubatus_core', 'jubatus_util_text'],
            language='c++',
            extra_compile_args=extra_compile_args)
    ]),
    install_requires=['jubatus'],
    test_suite='tests',
)



import json
import unittest

from embedded_jubatus import Anomaly
from jubatus.anomaly.types import IdWithScore
from jubatus.common import Datum


CONFIG = {
    'method': 'lof',
    'parameter': {
        'nearest_neighbor_num': 3,
        'reverse_nearest_neighbor_num': 5,
        'method': 'lsh',
        'parameter': {
            'hash_num': 512,
        }
    },
    'converter': {
        'num_filter_types': {},
        'num_filter_rules': [],
        'string_filter_types': {},
        'string_filter_rules': [],
        'num_types': {},
        'num_rules': [
            {'key': '*', 'type': 'num'}
        ],
        'string_types': {},
        'string_rules': [
            {'key': '*', 'type': 'space',
             'sample_weight': 'bin', 'global_weight': 'bin'}
        ]
    },
}


class TestAnomaly(unittest.TestCase):
    def test_invalid_configs(self):
        self.assertRaises(TypeError, Anomaly)
        self.assertRaises(ValueError, Anomaly, {})
        self.assertRaises(TypeError, Anomaly, {'method': 'hoge'})
        self.assertRaises(RuntimeError, Anomaly,
                          {'method': 'hoge', 'converter': {}})
        invalid_config = dict(CONFIG)
        invalid_config['method'] = 'hoge'
        self.assertRaises(RuntimeError, Anomaly, invalid_config)

    def test(self):
        x = Anomaly(CONFIG)
        self.assertEqual(float('inf'), x.add(Datum({'x': 0.1})).score)
        self.assertTrue(isinstance(x.add(Datum({'x': 0.101})), IdWithScore))
        x.add(Datum({'x': 0.1001}))
        x.calc_score(Datum({'x': 0.1001}))
        self.assertEqual(set(['0', '1', '2']), set(x.get_all_rows()))

        model = x.save_bytes()
        x.clear()
        self.assertEqual([], x.get_all_rows())

        x = Anomaly(CONFIG)
        x.load_bytes(model)
        self.assertEqual(set(['0', '1', '2']), set(x.get_all_rows()))
        self.assertEqual(CONFIG, json.loads(x.get_config()))

        p = x.add(Datum({'x': 0.2}))
        self.assertEqual(set(['0', '1', '2', '3']), set(x.get_all_rows()))
        self.assertEqual(p.score, x.update(p.id, Datum({'x': 0.2})))
        self.assertEqual(p.score, x.overwrite(p.id, Datum({'x': 0.2})))



cdef class _AnomalyWrapper:
    cdef _Anomaly *_handle

    def __cinit__(self):
        self._handle = NULL

    def __dealloc__(self):
        if self._handle != NULL:
            del self._handle

    def _init(self, config):
        self._handle = new _Anomaly(config)
        typ, ver = b'anomaly', 1
        return (
            lambda: self._handle.get_config().decode('utf8'),
            lambda: self._handle.dump(typ, ver),
            lambda x: self._handle.load(x, typ, ver),
            lambda: self._handle.clear(),
            typ
        )

    def clear_row(self, id_):
        self._handle.clear_row(id_.encode('utf8'))
        return True

    def add(self, row):
        cdef datum d
        cdef pair[string, float] r
        datum_py2native(row, d)
        r = self._handle.add(d)
        return AnomalyIdWithScore(r.first.decode('utf8'), r.second)

    def update(self, id_, row):
        cdef datum d
        datum_py2native(row, d)
        return self._handle.update(id_.encode('utf8'), d)

    def overwrite(self, id_, row):
        cdef datum d
        datum_py2native(row, d)
        return self._handle.overwrite(id_.encode('utf8'), d)

    def calc_score(self, row):
        cdef datum d
        datum_py2native(row, d)
        return self._handle.calc_score(d)

    def get_all_rows(self):
        cdef vector[string] ret = self._handle.get_all_rows()
        return [(<string>ret[i]).decode('utf8') for i in range(ret.size())]
